# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides model runner utilities
"""
from os.path import dirname, join as path_join


_DATA_DIR = path_join(dirname(__file__), 'data')


def data_path(*path, prefix=_DATA_DIR):
    '''Returns an absolute path given a relative path from the acumos_model_runner/data/ dir'''
    return path_join(prefix, *path)


def load_data(*path, prefix=_DATA_DIR, mode='r', loader=lambda file: file.read()):
    '''Loads and returns the contents of a file in the acumos_model_runner/data/ dir'''
    with open(data_path(*path, prefix=prefix), mode) as file:
        return loader(file)



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides utilities for parsing protobuf IDL
"""
from collections import namedtuple

from lark import Lark, Transformer

from acumos_model_runner.utils import load_data


Enum = namedtuple('Enum', 'name, enums')
Message = namedtuple('Message', 'name, fields, messages, enums')
Field = namedtuple('Field', 'type, name, number')
RepeatedField = namedtuple('RepeatedField', 'type, name, number')
MapField = namedtuple('MapField', 'key_type, val_type, name, number')

_PROTO_GRAMMAR = load_data('proto3.ebnf')


class ProtoTransformer(Transformer):
    '''Converts syntax tree token into more easily usable namedtuple objects'''

    def message(self, tokens):
        '''Returns a Message namedtuple'''
        name_token, body = tokens
        return Message(name_token.value, *body)

    def messagebody(self, items):
        '''Returns a tuple of message body namedtuples'''
        messages = []
        enums = []
        fields = []
        for item in items:
            if isinstance(item, Message):
                messages.append(item)
            elif isinstance(item, Enum):
                enums.append(item)
            elif isinstance(item, (Field, RepeatedField, MapField)):
                fields.append(item)
        return fields, messages, enums

    def field(self, tokens):
        '''Returns a Field namedtuple'''
        type_, fieldname, fieldnumber = (token.value for token in tokens)
        return Field(type_, fieldname, int(fieldnumber))

    def repeatedfield(self, field):
        '''Returns a RepeatedField namedtuple'''
        return RepeatedField(*field[0])

    def mapfield(self, tokens):
        '''Returns a MapField namedtuple'''
        key_type, val_type, fieldname, fieldnumber = (token.value for token in tokens)
        return MapField(key_type, val_type, fieldname, int(fieldnumber))

    def enum(self, tokens):
        '''Returns an Enum namedtuple'''
        name, enums = tokens
        return Enum(name.value, enums)

    def enumbody(self, tokens):
        '''Returns a sequence of enum identifiers'''
        enums = [enumfield.children[0].value for enumfield in tokens if enumfield.data == 'enumfield']
        return enums


def parse_proto(proto_idl):
    '''Returns a sequence of top-level protobuf definitions, i.e. Message or Enum namedtuples'''
    parser = Lark(_PROTO_GRAMMAR, start='proto', parser='lalr')
    tree = parser.parse(proto_idl)
    trans_tree = ProtoTransformer().transform(tree)
    top_level = [child for top_level in trans_tree.find_data('topleveldef')
                 for child in top_level.children if isinstance(child, (Message, Enum))]
    return top_level



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
"""
Provides utilities for generating an Open API specification from protobuf IDL and model metadata
"""
import yaml
from collections import namedtuple

from jinja2 import Environment, FileSystemLoader

from acumos_model_runner.proto_parser import Message, RepeatedField, MapField, Enum, parse_proto
from acumos_model_runner.utils import data_path


class TemplateError(Exception):
    pass


_OasFormat = namedtuple('_OasFormat', 'type, format, description')
_OasFormat.__new__.__defaults__ = (None, None, None)

_64NOTE = 'Note: 64-bit integers are string-encoded when using application/json'

_PROTO_OAS_MAP = {
    'double': _OasFormat('number', 'double'),
    'float': _OasFormat('number', 'float'),
    'int32': _OasFormat('integer', 'int32'),
    'int64': _OasFormat('integer', 'int64', _64NOTE),
    'uint32': _OasFormat('integer', 'int32'),
    'uint64': _OasFormat('integer', 'int64', _64NOTE),
    'sint32': _OasFormat('integer', 'int32'),
    'sint64': _OasFormat('integer', 'int64', _64NOTE),
    'fixed32': _OasFormat('integer', 'int32'),
    'fixed64': _OasFormat('integer', 'int64', _64NOTE),
    'sfixed32': _OasFormat('integer', 'int32'),
    'sfixed64': _OasFormat('integer', 'int64', _64NOTE),
    'bool': _OasFormat('boolean'),
    'string': _OasFormat('string'),
    'bytes': _OasFormat('string', 'byte')}


def create_oas(metadata, protobuf):
    '''Returns an OAS YAML string'''
    top_level = parse_proto(protobuf)
    defs = _create_definitions(top_level)

    defs_yaml = yaml.dump(defs, default_flow_style=False)
    methods = [_format_method(name, method) for name, method in metadata['methods'].items()]

    env = Environment(loader=FileSystemLoader(data_path('templates')), trim_blocks=True)
    return env.get_template('base.yaml').render(model=metadata, methods=methods, definitions=defs_yaml)


def _format_method(name, method):
    '''Returns a method dict to be used in the method template'''
    method_fmt = dict(name=name, **method)
    for key in ('input', 'output'):
        method_fmt[key] = _prefix_name(method[key])
    return method_fmt


def _create_definitions(top_level):
    '''Returns OAS definitions for all protobuf top-level definitions'''
    refs = {ref for item in top_level for ref in _find_refs(item)}
    defs = {name: definition for item in top_level for name, definition in _find_definitions(item, refs)}
    defs_prefixed = {_prefix_name(key): val for key, val in defs.items()}
    return defs_prefixed


def _find_refs(item, root=()):
    '''Yields tuples representing a hierarchy of custom named types'''
    yield root + (item.name, )
    if isinstance(item, Message):
        child_root = root + (item.name, )
        for enum in item.enums:
            yield child_root + (enum.name, )
        for msg in item.messages:
            yield from _find_refs(msg, child_root)


def _find_definitions(item, refs, prefix=()):
    '''Yields OAS definitions contained within a top-level protobuf definition'''
    if isinstance(item, Message):
        yield from _find_message_definitions(item, refs, prefix)
    elif isinstance(item, Enum):
        yield _define_enum(item, prefix)
    else:
        raise TemplateError("Cannot create definition item {}".format(item))


def _find_message_definitions(message, refs, prefix):
    '''Yields OAS definitions contained within to a protobuf message'''
    scope = prefix + (message.name, )

    for enum in message.enums:
        yield _define_enum(enum, scope)

    for msg in message.messages:
        yield from _find_message_definitions(msg, refs, scope)

    properties = {field.name: _define_field(field, refs, scope) for field in message.fields}

    def_obj = {'type': 'object', 'properties': properties, 'required': sorted(properties.keys())}
    def_name = ".".join(scope)
    yield def_name, def_obj


def _define_enum(enum, prefix):
    '''Returns an OAS object corresponding to a protobuf enum'''
    def_name = ".".join(prefix + (enum.name, ))
    return def_name, {'type': 'string', 'enum': enum.enums}


def _define_field(field, refs, prefix):
    '''Returns an OAS object corresponding to a protobuf field'''
    if isinstance(field, MapField):
        field_type = _resolve_type(field.val_type, refs, prefix)
        return {'type': 'object', 'additionalProperties': field_type}
    else:
        field_type = _resolve_type(field.type, refs, prefix)
        if isinstance(field, RepeatedField):
            return {'type': 'array', 'items': field_type}
        else:
            return field_type


def _resolve_type(type_name, refs, prefix):
    '''Returns an OAS object corresponding to a protobuf type. Returns a reference for named types'''
    if type_name in _PROTO_OAS_MAP:
        oas_type = _PROTO_OAS_MAP[type_name]
        return {k: v for k, v in oas_type._asdict().items() if v is not None}
    else:
        ref_tuple = _resolve_named_type(type_name, refs, prefix)
        ref_str = "#/definitions/{}".format(_prefix_name(".".join(ref_tuple)))
        return {'$ref': ref_str}


def _resolve_named_type(type_name, refs, prefix):
    '''Returns the full identifier for a given named type'''
    type_idents = tuple(type_name.split('.'))

    # look within the most local scope first, and gradually move towards global scope
    for i in reversed(range(len(prefix) + 1)):
        full_ident = prefix[:i] + type_idents
        if full_ident in refs:
            return full_ident
    else:
        raise TemplateError("Failed to find a reference for named type {}".format('/'.join(prefix + type_idents)))


def _prefix_name(name):
    '''Adds a "Model" prefix to type names to avoid naming collisions'''
    return "Model.{}".format(name)



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
'''
Provides model runner API implementations
'''
from functools import partial

from flask import current_app, send_from_directory, request, abort, Response
from google.protobuf.message import DecodeError
from google.protobuf.json_format import ParseError


_PROTO = 'application/vnd.google.protobuf'
_JSON = 'application/json'
_VALID_MEDIA = {_PROTO, _JSON}


def methods(method_name):
    '''Generic handler for model methods'''
    content_type = _get_header('Content-Type')
    accept = _get_header('Accept')

    data = request.data
    method = current_app.model.methods[method_name]

    try:
        wrapped_resp = method.from_pb_bytes(data) if content_type == _PROTO else method.from_json(data)
    except DecodeError as err:
        abort(Response("Could not decode input protobuf message: {}".format(err), 400))
    except ParseError as err:
        abort(Response("Could not parse input JSON message: {}".format(err), 400))
    except Exception as err:
        abort(Response("Could not invoke method due to runtime error: {}".format(err), 400))

    resp_data = wrapped_resp.as_pb_bytes() if accept == _PROTO else wrapped_resp.as_json()
    return Response(resp_data, status=200, content_type=accept)


def _get_header(name):
    '''Returns a given request header'''
    header = request.headers.get(name)
    if header is None:
        abort(Response("Header '{}' is required".format(name), 400))
    if header not in _VALID_MEDIA:
        abort(Response("Header '{}' must be one of {}".format(name, _VALID_MEDIA), 415))
    return header


def artifacts(filename):
    '''Generic handler for model artifacts'''
    return send_from_directory(current_app.model_dir, filename)


metadata = partial(artifacts, filename='metadata.json')
protobuf = partial(artifacts, filename='model.proto')



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
from os.path import dirname, abspath, join as path_join
from setuptools import setup, find_packages


SETUP_DIR = abspath(dirname(__file__))
DOCS_DIR = path_join(SETUP_DIR, 'docs')

with open(path_join(SETUP_DIR, 'acumos_model_runner', '_version.py')) as file:
    globals_dict = dict()
    exec(file.read(), globals_dict)
    __version__ = globals_dict['__version__']


def _long_descr():
    '''Yields the content of documentation files for the long description'''
    for file in ('user-guide.rst', path_join('tutorial', 'index.rst'), 'release-notes.rst', 'developer-guide.rst'):
        doc_path = path_join(DOCS_DIR, file)
        with open(doc_path) as f:
            yield f.read()


setup(
    author='Paul Triantafyllou, Eric Zavesky',
    author_email='trianta@research.att.com, ezavesky@research.att.com',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: Apache Software License',
    ],
    description='Acumos model runner for Python models',
    entry_points="""
    [console_scripts]
    acumos_model_runner=acumos_model_runner.runner:run_app_cli
    """,
    install_requires=['acumos>=0.5.3',
                      'lark-parser',
                      'connexion<2.0.0',
                      'gunicorn',
                      'pyyaml',
                      'jinja2',
                      'protobuf',
                      'flask-cors',
                      'swagger_ui_bundle'],
    keywords='acumos machine learning model runner server protobuf ml ai',
    license='Apache License 2.0',
    long_description='\n'.join(_long_descr()),
    name='acumos_model_runner',
    packages=find_packages(),
    package_data={'acumos_model_runner': [path_join('data', 'proto3.ebnf'),
                                          path_join('data', 'templates', '*.yaml')]},
    python_requires='>=3.4',
    url='https://gerrit.acumos.org/r/gitweb?p=python-model-runner.git',
    version=__version__,
)



# -*- coding: utf-8 -*-
# ===============LICENSE_START=======================================================
# Acumos Apache-2.0
# ===================================================================================
# Copyright (C) 2017-2018 AT&T Intellectual Property & Tech Mahindra. All rights reserved.
# ===================================================================================
# This Acumos software file is distributed by AT&T and Tech Mahindra
# under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# This file is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============LICENSE_END=========================================================
'''
Provides a model runner based on a connexion application and gunicorn server
'''
import json
import argparse
from functools import partial
from os.path import abspath, join as path_join

from gunicorn.app.base import BaseApplication
from connexion import App
from connexion.resolver import Resolver
from swagger_ui_bundle import swagger_ui_3_path as swagger_ui_path
from flask import redirect, Blueprint, render_template
from flask_cors import CORS
from acumos.wrapped import load_model

from acumos_model_runner.api import methods
from acumos_model_runner.oas_gen import create_oas


def run_app_cli():
    '''CLI entry point for starting the model runner'''
    parser = argparse.ArgumentParser()
    parser.add_argument('model_dir', type=str, help='Directory containing a dumped Acumos Python model')
    parser.add_argument('--host', type=str, default='0.0.0.0', help='The interface to bind to')
    parser.add_argument('--port', type=int, default=3330, help='The port to bind to')
    parser.add_argument('--workers', type=int, default=1, help='The number of gunicorn workers to spawn')
    parser.add_argument('--timeout', type=int, default=120, help='Time to wait (seconds) before a frozen worker is restarted')
    parser.add_argument('--cors', type=str, default=None, help="Enables CORS if provided. Can be a domain, comma-separated list of domains, or '*'")

    pargs = parser.parse_args()

    app = create_app(**vars(pargs))
    app.run()


def create_app(model_dir, host, port, workers=1, timeout=120, cors=None):
    '''Creates and returns the model runner gunicorn application

    Parameters
    ----------
    model_dir : str
        Directory containing a dumped Acumos Python model
    host : str
        The interface to bind to
    port : int
        The port to bind to
    workers : int, optional
        The number of gunicorn workers to spawn
    timeout : int, optional
        Time to wait (seconds) before a frozen worker is restarted
    cors : str, optional
        Enables CORS if provided. Can be a domain, comma-separated list of domains, or '*'
    '''
    _write_oas(model_dir)
    return StandaloneApplication(model_dir, host, port, workers, timeout, cors)


def _write_oas(model_dir):
    '''Writes an Open API specification file the model directory'''
    with open(path_join(model_dir, 'metadata.json')) as file:
        metadata = json.load(file)

    with open(path_join(model_dir, 'model.proto')) as file:
        proto = file.read()

    oas_yaml = create_oas(metadata, proto)
    with open(path_join(model_dir, 'oas.yaml'), 'w') as file:
        file.write(oas_yaml)


class StandaloneApplication(BaseApplication):
    '''Custom gunicorn app. Modified from http://docs.gunicorn.org/en/stable/custom.html'''

    def __init__(self, model_dir, host, port, workers, timeout, cors):
        self.model_dir = model_dir
        self.cors = cors
        self.options = {'bind': "{}:{}".format(host, port), 'workers': workers, 'timeout': timeout}
        super().__init__()

    def load_config(self):
        config = dict([(key, value) for key, value in self.options.items()
                       if key in self.cfg.settings and value is not None])
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return _build_app(self.model_dir, self.cors)


def _build_app(model_dir, cors):
    '''Builds and returns a Flask app'''
    connexion_app = App(__name__, specification_dir=abspath(model_dir), swagger_ui=False)
    connexion_app.add_api('oas.yaml', resolver=_CustomResolver())

    flask_app = connexion_app.app
    flask_app.model = load_model(model_dir)
    flask_app.model_dir = model_dir

    @flask_app.route('/')
    def redirect_ui():
        return redirect('/ui')

    _add_swagger(flask_app)
    _apply_cors(flask_app, cors)

    return flask_app


class _CustomResolver(Resolver):

    def resolve_function_from_operation_id(self, operation_id):
        '''Routes model methods to a generic handler so that methods can be enumerated in the OAS'''
        if operation_id.startswith('methods'):
            _, method_name = operation_id.split('.')
            return partial(methods, method_name=method_name)
        else:
            return super().resolve_function_from_operation_id(operation_id)


def _add_swagger(app):
    '''Manually adds a Swagger v3 UI to the Flask app (until connexion 2.0 is released)'''
    swagger_bp = Blueprint('swagger_ui', __name__, static_url_path='', static_folder=swagger_ui_path, template_folder=swagger_ui_path)

    @swagger_bp.route('/')
    def swagger_ui_index():
        return render_template('index.j2', openapi_spec_url='/swagger.json')

    app.register_blueprint(swagger_bp, url_prefix='/ui')


def _apply_cors(app, cors):
    '''Configures a Flask app with CORS'''
    if isinstance(cors, str):
        origins = cors if cors == '*' else cors.split(',')
        CORS(app, origins=origins)
